
uniform Vec2_h terrain_texture_origin;
uniform Vec2_h terrain_texture_size;
uniform Vec4_h terrain_heights; // x: min, y: max, z: delta, w: max_delta

uniform Vec3_h terrain_intersection_min;
uniform Vec3_h terrain_intersection_max;

varying Vec4_h color;
varying Vec4_h worldPosition;
varying Vec4_h worldNormal;
varying Vec2_h terrain_texture_uv;
varying Vec2_h island_color_texture_uv;

//------------------------------------------------------------------------------
#ifdef Vertex

uniform Mat4_h worldMatrix;

void main()
{
    color = colorUniform;

    worldPosition =
        MULT_MAT_VEC(
            worldMatrix,
            Vec4(vertexPosition.xyz, 1.0));

    worldNormal =
        normalize(
            MULT_MAT_VEC(
                worldMatrix,
                Vec4(vertexNormal.xyz, 0.0)));

    pixelPosition = MULT_MAT_VEC(viewProjMatrix, worldPosition);

    terrain_texture_uv =
        (worldPosition.xz - terrain_texture_origin) / terrain_texture_size;

    island_color_texture_uv = vertexUV0;
}

#endif // Vertex
//------------------------------------------------------------------------------
#ifdef Pixel

Vec3_h hue_to_rgb(Float_h H)
{
    Float_h r = abs(H * 6.0 - 3.0) - 1.0;
    Float_h g = 2.0 - abs(H * 6.0 - 2.0);
    Float_h b = 2.0 - abs(H * 6.0 - 4.0);

    return clamp(Vec3(r, g, b), Vec3(0.0), Vec3(1.0));
}

const Float_h EPSILON = 1e-10;

Vec3_h rgb_to_hcv(Vec3_h RGB)
{
    Vec4_h P = (RGB.g < RGB.b) ? Vec4(RGB.bg, -1.0, 2.0/3.0) : Vec4(RGB.gb, 0.0, -1.0/3.0);
    Vec4_h Q = (RGB.r < P.x) ? Vec4(P.xyw, RGB.r) : Vec4(RGB.r, P.yzx);

    Float_h C = Q.x - min(Q.w, Q.y);
    Float_h H = abs((Q.w - Q.y) / (6.0 * C + EPSILON) + Q.z);

    return Vec3(H, C, Q.x);
}

Vec3_h rgb_to_hsl(Vec3_h rgb)
{
    Vec3_h hcv = rgb_to_hcv(rgb);

    Float_h L = hcv.z - hcv.y * 0.5;
    Float_h S = hcv.y / (1.0 - abs(L * 2.0 - 1.0) + EPSILON);

    return Vec3(hcv.x, S, L);
}

Vec3_h hsl_to_rgb(Vec3_h HSL)
{
    Vec3_h RGB = hue_to_rgb(HSL.x);

    Float_h C = (1.0 - abs(2.0 * HSL.z - 1.0)) * HSL.y;

    return (RGB - 0.5) * C + HSL.z;
}

Vec3_h pastelize(Vec3_h rgb)
{
    Vec3_h hsl = rgb_to_hsl(rgb);

    hsl.z = 0.825;

    return hsl_to_rgb(hsl);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Vec4_h black  = Vec4(0.0, 0.0, 0.0, 1.0);
Vec4_h white  = Vec4(1.0, 1.0, 1.0, 1.0);
Vec4_h yellow = Vec4(1.0, 1.0, 0.0, 1.0);

const Float_h PI   = 3.14159265359;
const Float_h PI_2 = 6.28318530718;

void main()
{
    Vec4_h terrain_texel =
            FETCH_TEXEL_2D(terrain_texture, terrain_texture_uv);

    Vec4_h grid_color = black;

    outputColor = Vec4(0.0, 0.0, 0.0, 0.0);

    Float_h up = worldNormal.y;

    if (up < 0.5)
        return;

    Float_h pixel_height = worldPosition.y;

    Float_h texel_height_normalized = terrain_texel.a;

    Float_h terrain_height_min       = terrain_heights.x;
    Float_h terrain_height_delta     = terrain_heights.z;
    Float_h terrain_height_max_delta = terrain_heights.w;

    Float_h texel_height = (texel_height_normalized * terrain_height_delta) + terrain_height_min;
    Float_h texel_height_min = texel_height - terrain_height_max_delta;
    Float_h texel_height_max = texel_height + terrain_height_max_delta;

    bool pixel_height_is_in_texel_height_range =
        (pixel_height > texel_height_min) &&
        (pixel_height < texel_height_max);

    if (!pixel_height_is_in_texel_height_range)
        return;

    bool is_obstructed = terrain_texel.r > 0.0;

    if (is_obstructed)
    {
        int encoded_obstacles = int(terrain_texel.r * 255.0);

        if (encoded_obstacles == 2)
        {
            // obstructed by barriers only - black
            outputColor = Vec4(0.0, 0.0, 0.0, 1.0);

            grid_color = white;
        }
        else
        {
            // obstructed by terrain slope - no color
            outputColor = Vec4(0.0, 0.0, 0.0, 0.0);
        }
    }
    else
    {
        Float_h island_color_index = terrain_texel.g;

        Vec2_h island_color_uv = Vec2(island_color_index, 0.5);

        Vec4_h island_color =
            FETCH_TEXEL_2D(island_color_texture, island_color_texture_uv);

        outputColor.rgb = island_color.rgb;

        // pastelize island color
        //outputColor.rgb = (outputColor.rgb + Vec3(1.2)) / Vec3(2.0);

        outputColor.rgb = pastelize(outputColor.rgb);

        outputColor.a = 0.8;
    }

    bool is_on_path = terrain_texel.b > 0.0;

    if (is_on_path)
    {
        outputColor.rgb = (outputColor.rgb + white.rgb) / Vec3(2.0);
    }

    Vec3_h grid_interval = Vec3(1.0);

    Vec3_h p = worldPosition.xyz / grid_interval;

    // distance from camera
    Float_h d       = length(p - eye);
    Float_h alpha_d = 1.0 - smoothstep(0.0, 50.0, d);

    Vec3_h one = Vec3(1.0);
    Vec3_h two = Vec3(2.0);

    // distance from grid, [0..0.5]
    Vec3_h from_grid = max(fract(p), fract(one - p));

    // nearness to grid [0..1]
    Vec3_h near_grid = one - (from_grid * 2.0);

    Vec3_h pow_near_grid = pow(near_grid, Vec3(20.0));

    Float_h on_grid_edge = max(pow_near_grid.x, pow_near_grid.z);

    // tile alignment
    Vec3_h floor_p = floor(p);

    Vec3_h floor_p_mod2 = mod(floor_p, two);

    Float_h in_odd_tile = abs(floor_p_mod2.x - floor_p_mod2.z);

    Float_h alpha
        = (on_grid_edge + in_odd_tile * 0.15)
        * alpha_d
        * 0.4;

    // Float_h inv_alpha = 1.0 - alpha;

    outputColor.rgb = mix(outputColor.rgb, grid_color.rgb, alpha);
        // outputColor.rgb * Vec3(inv_alpha) +
        //  grid_color.rgb * Vec3(    alpha);

    bool pixel_is_in_terrain_intersection =
        all(greaterThanEqual(worldPosition.xyz, terrain_intersection_min)) &&
        all(   lessThanEqual(worldPosition.xyz, terrain_intersection_max));

    if (pixel_is_in_terrain_intersection)
    {
        Vec3_h pow_near_grid = pow(near_grid, Vec3(10.0));

        Float_h on_grid_edge = max(pow_near_grid.x, pow_near_grid.z);

        outputColor.rgb = mix(outputColor.rgb, grid_color.rgb * Vec3(5.0), on_grid_edge);
    }
}

#endif // Pixel